"""nova stacks controller."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from cement.core.controller import CementBaseController, expose
import pip


class NovaUpdateController(CementBaseController):
    class Meta:
        label = 'update'
        stacked_on = 'base'
        stacked_type = 'nested'
        description = 'Update NOVA to latest version'
        usage = "nova update"

    @expose(hide=True)
    def default(self):
        print("updating NOVA, please wait...")
        # TODO Update this once open-sourced to PyPi
        pip.main(['install', '--upgrade', 'gilt-nova'])



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
import os
import hashlib
from botocore.exceptions import ClientError
from termcolor import colored

from nova.core.exc import NovaError


def create_and_upload_stack_template(s3, s3_bucket, service, environment):
    bucket=None
    try:
        bucket = s3.head_bucket(Bucket=s3_bucket)
    except ClientError:
        print(colored('Common Nova service templates bucket does not exist, creating...', color='cyan'))
        s3.create_bucket(Bucket=s3_bucket)

    for s, template in environment.get_stack_templates_used(service, s3_bucket).items():
        if 's3_key' in template:
            template_file = template['filename']
            template_s3_key = template['s3_key']
        else:
            template_version=template['version']
            template_filename='%s.json' % template['name']
            user_home_dir = os.path.join(os.path.expanduser('~'), '.nova')

            if not os.path.exists(user_home_dir):
                raise NovaError("Unable to find NOVA templates. Please create templates under '~/.nova'.")

            template_file = os.path.join(user_home_dir, template_version, template_filename)
            template_s3_key = '%s/%s' % (template_version, template_filename)

        existing_file=None
        if bucket:
            print(colored('Found common Nova service templates bucket. Verifying contents...', color='cyan'))
            checksum = md5(template_file)
            try:
                existing_file = s3.head_object(Bucket=s3_bucket, Key=template_s3_key, IfMatch=checksum)
            except ClientError:
                pass

        if not existing_file:
            print(colored('Common Nova service templates not found. Uploading...', color='cyan'))
            with open(template_file, 'rb') as template_data:
                s3.put_object(Bucket=s3_bucket, Key=template_s3_key, Body=template_data)
            print(colored('Contents uploaded!', color='green'))
        else:
            print(colored('Contents verified!', color='green'))

    print(colored('All required templates have been uploaded & verified!', color='green'))


def md5(fname):
    md5_hash = hashlib.md5()
    with open(fname, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            md5_hash.update(chunk)
    return md5_hash.hexdigest()



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from nova.core import check_latest_version
from nova.core.exc import NovaError
from boto3.session import Session
from botocore.exceptions import ClientError
from base64 import b64decode
from Crypto.Cipher import AES
from Crypto.Hash import SHA256
from Crypto.Hash.HMAC import HMAC
from Crypto.Util import Counter


class Decrypt:

    def __init__(self, stash_key, aws_profile=None, aws_region=None, aws_bucket=None, context=None):
        check_latest_version()

        if not context:
            context = {}

        session = Session(profile_name=aws_profile, region_name=aws_region)

        account_id = session.client('iam').list_account_aliases()['AccountAliases'][0]
        deployment_bucket_name = aws_bucket or 'novastash_%s' % account_id

        s3 = session.client('s3')
        kms = session.client('kms')

        key = "%s.txt.enc" % stash_key
        existing_stash = None
        try:
            existing_stash = s3.get_object(Bucket=deployment_bucket_name, Key=key)
        except ClientError as e:
            pass

        if existing_stash is None:
            raise NovaError("No stash '%s' found!" % stash_key)
        else:
            contents = existing_stash['Body'].read()
            metadata = existing_stash['Metadata']
            encryption_key = metadata['encryption-key']
            try:
                kms_response = kms.decrypt(CiphertextBlob=b64decode(encryption_key), EncryptionContext=context)
            except ClientError as e:
                if e.response["Error"]["Code"] == "InvalidCiphertextException":
                    if context is None:
                        msg = ("Could not decrypt hmac key with KMS. The credential may "
                               "require that an encryption context be provided to decrypt it.")
                    else:
                        msg = ("Could not decrypt hmac key with KMS. The encryption "
                               "context provided may not match the one used when the "
                               "credential was stored.")
                else:
                    msg = "Decryption error %s" % e
                raise NovaError(msg)
            except Exception as e:
                raise NovaError("Decryption error %s" % e)

            key = kms_response['Plaintext'][:32]
            hmac_key = kms_response['Plaintext'][32:]
            hmac = HMAC(hmac_key, msg=b64decode(contents), digestmod=SHA256)
            if hmac.hexdigest() != metadata['hmac']:
                raise NovaError("Computed HMAC on '%s' does not match stored HMAC" % stash_key)
            dec_ctr = Counter.new(128)
            decryptor = AES.new(key, AES.MODE_CTR, counter=dec_ctr)
            value = decryptor.decrypt(b64decode(contents)).decode("utf-8")
            print(value)


from botocore.exceptions import ClientError


def check_stash_exists(s3, deployment_bucket_name, key):
    existing_obj=None
    try:
        existing_obj = s3.head_object(Bucket=deployment_bucket_name, Key=key)
    except ClientError:
        pass

    return existing_obj


def kms_key_exists(kms, kms_key):
    return kms.describe_key(KeyId=kms_key) is not None



from setuptools import setup, find_packages
from pip.req import parse_requirements
import os
import glob

conf = []

for name in glob.glob('config/plugins.d/*.conf'):
    conf.insert(1, name)

if not os.path.exists('~/.nova/'):
    os.makedirs('~/.nova/')

# parse_requirements() returns generator of pip.req.InstallRequirement objects
install_reqs = parse_requirements('requirements.txt', session=False)

setup(name='gilt-nova',
    version='10.5.2',
    description="Collection of utilities to easily deploy services to AWS.",
    long_description="Collection of utilities to easily deploy services to AWS.",
    license='Apache License 2.0',
    classifiers=[
        'Development Status :: 4 - Beta',

        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',

        'License :: OSI Approved :: Apache Software License',

        'Operating System :: POSIX :: Linux',

        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
    ],
    keywords='AWS CodeDeploy Cloudformation Deployment Gilt',
    author='Gilt Groupe',
    author_email='grhodes@gilt.com',
    url='https://github.com/gilt/nova',
    packages=find_packages(exclude=['ez_setup', 'contrib', 'docs']),
    include_package_data=True,
    zip_safe=False,
    install_requires=[str(ir.req) for ir in install_reqs],
    data_files=[('~/.nova', ['config/nova.conf']),
                ('~/.nova/plugins.d', conf)],
    setup_requires=[],
    entry_points="""
        [console_scripts]
        nova = nova.cli.main:main
    """,
    namespace_packages=[],
)




"""nova stash controller."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from cement.core.controller import CementBaseController, expose
from nova.core.stash.decrypt import Decrypt
from nova.core.stash.encrypt import Encrypt
from nova.core.exc import NovaError


class NovaStashController(CementBaseController):
    class Meta:
        label = 'stash'
        description = 'NOVA service stash tools'
        stacked_on = 'base'
        stacked_type = 'nested'
        arguments = [
            (['-p', '--profile'], dict(help='AWS profile')),
            (['-r', '--region'], dict(help='AWS region')),
            (['-b', '--bucket'], dict(help='AWS bucket')),
            (['stash_args'], dict(action='store', nargs='*')),
        ]

    @expose(hide=True)
    def default(self):
        self.app.args.print_help()

    @expose(help='Decrypt a service variable')
    def get(self):
        if len(self.app.pargs.stash_args) == 1:
            profile = self.app.pargs.profile
            region = self.app.pargs.region
            bucket = self.app.pargs.bucket
            key = self.app.pargs.stash_args[0]
            Decrypt(key, profile, region, bucket)
        else:
            raise NovaError("Usage: nova stash get <environment> <key>")

    @expose(help='Encrypt a service variable')
    def put(self):
        if len(self.app.pargs.stash_args) == 2:
            profile = self.app.pargs.profile
            region = self.app.pargs.region
            bucket = self.app.pargs.bucket
            key = self.app.pargs.stash_args[0]
            value = self.app.pargs.stash_args[1]
            Encrypt(key, value, profile, region, bucket)
        else:
            raise NovaError("Usage: nova stash put <environment> <key> <value>")



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from nova.core import query_yes_no, check_latest_version
from nova.core.stash import check_stash_exists, kms_key_exists
from nova.core.exc import NovaError
from boto3.session import Session
from botocore.exceptions import ClientError
from base64 import b64encode
from Crypto.Cipher import AES
from Crypto.Hash import SHA256
from Crypto.Hash.HMAC import HMAC
from Crypto.Util import Counter
from termcolor import colored
import sys


class Encrypt:

    def __init__(self, stash_key, value, aws_profile=None, aws_region=None, aws_bucket=None, kms_key="alias/novastash", context=None):
        check_latest_version()

        if not context:
            context = {}

        session = Session(profile_name=aws_profile, region_name=aws_region)

        account_id = session.client('iam').list_account_aliases()['AccountAliases'][0]
        deployment_bucket_name = aws_bucket or 'novastash_%s' % account_id

        s3 = session.client('s3')
        kms = session.client('kms')

        if not kms_key_exists(kms, kms_key):
            raise NovaError("Please setup the novastash KMS key.")

        try:
            s3.head_bucket(Bucket=deployment_bucket_name)
        except ClientError:
            print(colored("Creating novastash bucket '%s'" % deployment_bucket_name))
            s3.create_bucket(Bucket=deployment_bucket_name)

        # generate a a 64 byte key.
        # Half will be for data encryption, the other half for HMAC
        try:
            kms_response = kms.generate_data_key(KeyId=kms_key, EncryptionContext=context, NumberOfBytes=64)
        except:
            raise NovaError("Could not generate key using KMS key %s: %s" % (kms_key, str(sys.exc_info()[0])))
        data_key = kms_response['Plaintext'][:32]
        hmac_key = kms_response['Plaintext'][32:]
        wrapped_key = kms_response['CiphertextBlob']

        enc_ctr = Counter.new(128)
        encryptor = AES.new(data_key, AES.MODE_CTR, counter=enc_ctr)

        c_text = encryptor.encrypt(value)
        # compute an HMAC using the hmac key and the ciphertext
        hmac = HMAC(hmac_key, msg=c_text, digestmod=SHA256)
        b64hmac = hmac.hexdigest()

        key = "%s.txt.enc" % stash_key
        existing_stash = check_stash_exists(s3, deployment_bucket_name, key)

        if existing_stash is None:
            print(colored("Stashing '%s'" % stash_key))
            s3.put_object(
                Bucket=deployment_bucket_name,
                Body=b64encode(c_text).decode('utf-8'),
                Key=key,
                Metadata={
                    'encryption-key': b64encode(wrapped_key).decode('utf-8'),
                    'hmac': b64hmac
                }
            )
        else:
            perform_overwrite = query_yes_no("Stash '%s' already exists, want to overwrite?" % stash_key, default="no")
            if perform_overwrite:
                s3.put_object(
                    Bucket=deployment_bucket_name,
                    Body=b64encode(c_text).decode('utf-8'),
                    Key=key,
                    Metadata={
                        'encryption-key': b64encode(wrapped_key).decode('utf-8'),
                        'hmac': b64hmac
                    }
                )
            else:
                print(colored("Not stashing anything for key '%s'" % stash_key))



"""nova stacks controller."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from cement.core.controller import CementBaseController, expose
from nova.core.exc import NovaError
from nova.core.stack.create_stack import CreateStack
from nova.core.stack.update_stack import UpdateStack


class NovaStacksController(CementBaseController):
    class Meta:
        label = 'stack'
        description = 'NOVA service stack tools'
        stacked_on = 'base'
        stacked_type = 'nested'
        arguments = [
            (['-p', '--profile'], dict(help='Override nova.yml AWS profile')),
            (['environment'], dict(action='store', nargs='*'))
        ]
        usage = "nova stack [create|update] <environment>"

    @expose(hide=True)
    def default(self):
        self.app.args.print_help()

    @expose(help='Create NOVA service stack')
    def create(self):
        if self.app.pargs.environment:
            profile = self.app.pargs.profile
            CreateStack(profile, self.app.pargs.environment[0])
        else:
            raise NovaError("You must provide an environment to create")

    @expose(help='Update NOVA service stack')
    def update(self):
        if self.app.pargs.environment:
            profile = self.app.pargs.profile
            UpdateStack(profile, self.app.pargs.environment[0])
        else:
            raise NovaError("You must provide an environment to update")
